from .logger import logger
from .cfg import Config
from datetime import datetime, timedelta
from contextlib import contextmanager
import json

class Queue():
    def __init__(self, db, name = Config.default_name):
        self._db = db
        self._name = name

    def _consume(self, id):
        self._db.execute(lambda o: f"""
            DELETE FROM {Config.table}
            WHERE id = {o(id)}
        """)
        logger.debug(f"Message: {id} consumed")

    def _get(self):
        logger.debug(f"Queue: {self._name} attempting to dequeue message.")
        now_dt = datetime.utcnow()
        cutoff_dt = now_dt - timedelta(seconds = Config.unlock_seconds)
        row = self._db.fetch_one(lambda o: f"""
            UPDATE {Config.table} 
            SET lock_dt = {o(now_dt)},
            retries = retries - 1
            WHERE id = (
                SELECT id FROM {Config.table}
                WHERE name = {o(self._name)}
                AND (lock_dt IS NULL OR lock_dt < {o(cutoff_dt)})
                AND retries > 0
                ORDER BY lock_dt NULLS FIRST
                FOR UPDATE SKIP LOCKED
                LIMIT 1
            ) RETURNING id, payload
        """)
        if row is None:
            logger.debug(f"Queue: {self._name} is empty and no message can be dequeued.")
            return None, None
        id, payload = row
        logger.debug(f"Queue: {self._name} has dequeued message: {id} with payload: {json.dumps(payload)}")
        return id, payload

    def put(self, payload, *, retries = 1):
        id, = self._db.fetch_one(lambda o: f"""
            INSERT INTO {Config.table} VALUES (
                DEFAULT,
                {o(self._name)},
                {o(json.dumps(payload))},
                NULL,
                {o(retries)}
            ) RETURNING id;
        """)
        logger.debug(f"Queue: {self._name} enqueued with message: {id} with payload: {json.dumps(payload)}.")
    
    @contextmanager
    def get(self):
        id, payload = self._get()
        if id is not None:
            yield payload, True
            self._consume(id)
        else:
            yield None, False

    @staticmethod
    def destroy_expired_messages(db):
        db.execute(lambda o: f"""
            DELETE FROM {Config.table}
            WHERE retries <= 0
        """)

    @staticmethod
    def setup(db):
        db.execute(lambda o: f"""
            CREATE TABLE IF NOT EXISTS {Config.table} (
                id SERIAL,
                name TEXT NOT NULL,
                payload JSONB NOT NULL,
                lock_dt TIMESTAMP NULL,
                retries INTEGER NOT NULL,
                PRIMARY KEY(id)
            );
        """)
        db.execute(lambda o: f"""
            CREATE INDEX IF NOT EXISTS {Config.table}_name_ix
                ON {Config.table}(name);
        """)
        db.execute(lambda o: f"""
            CREATE INDEX IF NOT EXISTS {Config.table}_lock_dt_ix
                ON {Config.table}(lock_dt NULLS FIRST);
        """)